import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
from numpy import save
import time
import itertools
import pandas as pd
import csv



from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout, Activation, Flatten, Conv2D, MaxPool2D, GlobalAveragePooling2D, AveragePooling2D, BatchNormalization
from tensorflow.keras import optimizers
from tensorflow.keras.optimizers import RMSprop
from tensorflow.keras.callbacks import ReduceLROnPlateau,EarlyStopping
from tensorflow.keras.regularizers import l2

from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix

print("Libraries loaded")

batch_size = 128
epochs = 10
num_classes = 10
weight_decay = 1e-6

#tf.enable_eager_execution() 
mnist = tf.keras.datasets.mnist

(x_train, y_train), (x_test, y_test) = mnist.load_data()
x_train, x_test = x_train / 255.0, x_test / 255.0


print(x_train.shape)
print(y_train.shape)
print(x_test.shape)
print(y_test.shape)

x_train = np.reshape(x_train,(-1,28,28,1))
x_test = np.reshape(x_test,(-1,28,28,1))
y_train = tf.keras.utils.to_categorical(y_train,10)
y_test = tf.keras.utils.to_categorical(y_test,10)


print(x_train.shape)
print(x_test.shape)
print(y_train.shape)
print(y_test.shape)





# un comment the BatchNormalization for datasets other than MNIST


model = Sequential()
    
model.add(Conv2D(192,(5,5),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal',input_shape=(28,28,1)))
#model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(160,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
#model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(96,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
#model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(MaxPool2D(pool_size=(3,3),strides=(2,2),padding='same'))
    
model.add(Dropout(0.2))
    
model.add(Conv2D(192,(5,5),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
#model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    #model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    #model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(MaxPool2D(pool_size=(3,3),strides=(2,2),padding='same'))
    
model.add(Dropout(0.2))
    
model.add(Conv2D(192,(3,3),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    #model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    #model.add(BatchNormalization())
model.add(Activation('relu'))
model.add(Conv2D(10,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    #model.add(BatchNormalization())
model.add(Activation('relu'))  
    
model.add(GlobalAveragePooling2D())
model.add(Activation('softmax'))

model.summary()
    
adam = optimizers.RMSprop()

model.compile(loss='categorical_crossentropy', optimizer=adam, metrics=['accuracy'])

print("Model is compiled.")

#____________________________________________________________________________________________________________________________

    
def convert(seconds): 
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

#____________________________________________________________________________________________________________________________

learning_rate_reduction = ReduceLROnPlateau(monitor='accuracy',patience=3,verbose=1,factor=0.1,min_lr = 0.001000000000101)

earlystopping = EarlyStopping(monitor='val_loss',patience=3,verbose=1,mode='auto')

start = time.time()

history = model.fit(x_train, y_train, epochs=10, callbacks=[learning_rate_reduction])

end = time.time()
total_time = end-start 
total_time = convert(total_time)
print("The total time to train the model was:", total_time)


with open('output.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(model.history.history['accuracy'])

    
la = []

loss, accuracy = model.evaluate(x_test, y_test)

print("Training Runtime: ", total_time)
print("The Test Set Loss is: ", loss, "\nThe Test Set Accuracy is: ", accuracy)

la.append(loss)
la.append(accuracy)
la.append(total_time)

print("The Test Error might be:", 1-accuracy)

with open('mnist_test_results.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(la)
    
#____________________________________________________________________________________________________________________________


Y_pred_test = np.zeros_like(y_test)
Y_pred_test = model.predict(x_test,batch_size=128)
Y_pred_test_classes = np.argmax(Y_pred_test,axis = 1) 
Y_true_test = np.argmax(y_test,axis = 1) 


errors = (Y_pred_test_classes != Y_true_test)

#pred labels
Y_pred_test_classes_errors = Y_pred_test_classes[errors]

#pred labels
Y_pred_test_errors = Y_pred_test[errors]

#true labels
Y_true_test_errors = Y_true_test[errors]

#images
X_val_test_errors = x_test[errors]


Y_pred_test_errors_prob = np.max(Y_pred_test_errors,axis = 1)

true_test_prob_errors = np.diagonal(np.take(Y_pred_test_errors, Y_true_test_errors, axis=1))

delta_pred_test_true_errors = Y_pred_test_errors_prob - true_test_prob_errors

sorted_dela_test_errors = np.argsort(delta_pred_test_true_errors)

most_important_test_errors = sorted_dela_test_errors[-6:]


save('mnist_img_test_errors.npy', X_val_test_errors)

with open('mnist_pred_test_errors.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_pred_test_classes_errors)

with open('mnist_true_test_errors.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_true_test_errors)

with open('mnist_err_test_index.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(most_important_test_errors)
    
    
with open('mnist_y_pred_test_classes.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_pred_test_classes)
    
with open('mnist_y_true_test.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_true_test)



#____________________________________________________________________________________________________________________________
