import tensorflow as tf
import matplotlib.pyplot as plt
import numpy as np
from numpy import save
import time
import itertools
import pandas as pd
import csv



from tensorflow.keras.preprocessing.image import ImageDataGenerator
from tensorflow.keras.models import Sequential
from tensorflow.keras.layers import Dense, Dropout, Activation, Flatten, Conv2D, MaxPool2D, GlobalAveragePooling2D, AveragePooling2D, BatchNormalization
from tensorflow.keras import optimizers
from tensorflow.keras.optimizers import RMSprop
from tensorflow.keras.callbacks import ReduceLROnPlateau,EarlyStopping
from tensorflow.keras.regularizers import l2

from sklearn.model_selection import train_test_split
from sklearn.metrics import confusion_matrix

print("Libraries loaded")

batch_size = 128
epochs = 30
num_classes = 10
weight_decay = 1e-6
#originally 15
nets = 15

train = pd.read_csv('./digit-recognizer/train.csv')
test = pd.read_csv('./digit-recognizer/test.csv')

labels = train['label']
train = train.drop(['label'],axis=1)
X = train.values
test = test.values

print(X.shape)
print(labels.shape)
print(test.shape)

X = X/255.
test = test/255.
X = np.reshape(X,(-1,28,28,1))
test = np.reshape(test,(-1,28,28,1))
labels = tf.keras.utils.to_categorical(labels,10)

print(X.shape)
print(test.shape)
print(labels.shape)

learning_rate_reduction = ReduceLROnPlateau(monitor='loss',
                                            patience=3,
                                            verbose=1,
                                            factor=0.5,
                                            min_delta=1e-5)
earlystopping = EarlyStopping(monitor='val_loss',
                              patience=3,
                              verbose=1,
                              mode='auto')


def build_model():
    model = Sequential()
    
    model.add(Conv2D(192,(5,5),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal',input_shape=(28,28,1)))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(160,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(96,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(MaxPool2D(pool_size=(3,3),strides=(2,2),padding='same'))
    
    model.add(Dropout(0.2))
    
    model.add(Conv2D(192,(5,5),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(MaxPool2D(pool_size=(3,3),strides=(2,2),padding='same'))
    
    model.add(Dropout(0.2))
    
    model.add(Conv2D(192,(3,3),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(192,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))
    model.add(Conv2D(10,(1,1),padding='same',kernel_regularizer=l2(weight_decay),kernel_initializer='he_normal'))
    model.add(BatchNormalization())
    model.add(Activation('relu'))  
    
    model.add(GlobalAveragePooling2D())
    model.add(Activation('softmax'))
    
    adam = optimizers.RMSprop()
    model.compile(loss='categorical_crossentropy',optimizer=adam,metrics=['accuracy'])
    return model

models = []
for i in range(nets):
    models.append(build_model())
    
def convert(seconds): 
    return time.strftime("%H:%M:%S", time.gmtime(seconds))

    
print('Using real-time data augmentation.')
start = time.time()
datagen = ImageDataGenerator(horizontal_flip=False,
                             width_shift_range=0.10,
                             height_shift_range=0.10,
                             fill_mode='constant',
                             cval=0,
                             rotation_range=10,
                             zoom_range=0.1)
histories = []
for i in range(nets):
    x_train,x_valid,y_train,y_valid = train_test_split(X,labels,test_size=0.2)
    datagen.fit(x_train)
    history = models[i].fit_generator(datagen.flow(x_train,y_train,batch_size=batch_size),
                                  steps_per_epoch=int(len(x_train)/batch_size),
                                  epochs=epochs,
                                  callbacks=[learning_rate_reduction,earlystopping],
                                  validation_data=(x_valid,y_valid),
                                  verbose=1)
    histories.append(history)
    print('Network in Network {0:d}: Epochs={1:d}, Train Accuracy={2:.5f}, Validation Accuracy={3:.5f}'.format(i+1,max(history.epoch),max(history.history['accuracy']),max(history.history['val_accuracy'])))
end = time.time()
total_time = end-start 
total_time = convert(total_time)
print("The total time to train the model was:", total_time)

loss_lst = []
val_loss_lst = []
acc_lst = []
val_acc_lst = []

for i in range(nets):
    history=histories[i]
    epoch_range = 1 + np.arange(len(history.history['accuracy']))
    loss_lst.append(history.history['loss'])
    val_loss_lst.append(history.history['val_loss'])
    acc_lst.append(history.history['accuracy'])
    val_acc_lst.append(history.history['val_accuracy'])

    

with open('kaggle_accuracy.csv','w') as file:
    writer = csv.writer(file)
    for word in acc_lst:
        writer.writerow(word)
        
with open('kaggle_val_accuracy.csv','w') as file:
    writer = csv.writer(file)
    for word in val_acc_lst:
        writer.writerow(word)
        
with open('kaggle_loss.csv','w') as file:
    writer = csv.writer(file)
    for word in loss_lst:
        writer.writerow(word)
        
with open('kaggle_val_loss.csv','w') as file:
    writer = csv.writer(file)
    for word in val_loss_lst:
        writer.writerow(word)
    
    
#____________________________________________________________________________________________________________________________

# For the Confusion Matrix    
    
    
Y_pred = np.zeros_like(y_valid)
for i in range(nets):
    model = models[i]
    Y_pred += model.predict(x_valid,batch_size=128)
Y_pred_classes = np.argmax(Y_pred,axis = 1) 
Y_true = np.argmax(y_valid,axis = 1) 

with open('kaggle_y_pred_classes.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_pred_classes)
    
with open('kaggle_y_true.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_true)
    
#____________________________________________________________________________________________________________________________

#For the errors

errors = (Y_pred_classes != Y_true)

#pred labels
Y_pred_classes_errors = Y_pred_classes[errors]

#pred labels
Y_pred_errors = Y_pred[errors]

#true labels
Y_true_errors = Y_true[errors]

#images
X_val_errors = x_valid[errors]


Y_pred_errors_prob = np.max(Y_pred_errors,axis = 1)

true_prob_errors = np.diagonal(np.take(Y_pred_errors, Y_true_errors, axis=1))

delta_pred_true_errors = Y_pred_errors_prob - true_prob_errors

sorted_dela_errors = np.argsort(delta_pred_true_errors)

most_important_errors = sorted_dela_errors[-6:]

    
save('kaggle_img_errors.npy', X_val_errors)

with open('kaggle_pred_errors.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_pred_classes_errors)

with open('kaggle_true_errors.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(Y_true_errors)

with open('kaggle_err_index.csv','w') as file:
    writer = csv.writer(file)
    writer.writerow(most_important_errors)

