#!/usr/bin/env bash
set -u

if [ $AVAILABLE_DEVICE == 'CPU' ]
then
    NODE_TYPE="${NODE_TYPE:-compute_haswell}"
else
    NODE_TYPE="${NODE_TYPE:-gpu_p100}"
fi

lease_name="$RESOURCE_NAME"
public_network_id=$(openstack network show public -f value -c id)

echo "Requesting lease for:"
echo " - 1 bare metal node (type=$NODE_TYPE)"
echo " - 1 Floating IP"
echo

blazar lease-create \
  --physical-reservation min=1,max=1,resource_properties="[\"=\", \"\$node_type\", \"$NODE_TYPE\"]" \
  --reservation resource_type=virtual:floatingip,network_id="$public_network_id",amount=1 \
  --start-date "$(date -u +'%Y-%m-%d %H:%M')" \
  --end-date "$(date -u +'%Y-%m-%d %H:%M' -d'+1 day')" \
  "$lease_name"

echo "Waiting for lease to start"
echo
attempts=0
wait_lease() {
  sleep 5
  attempts=$((attempts+1))
  printf "."
  [[ $attempts -gt 60 ]] && return 1
}
until [[ $(blazar lease-show "$lease_name" -f value -c status) == "ACTIVE" ]]; do wait_lease; done
echo
echo "Lease started successfully!"

reservations=$(blazar lease-show "$lease_name" -f json \
  | jq -r '.reservations')
host_reservation_id=$(jq -rs 'map(select(.resource_type=="physical:host"))[].id' <<<"$reservations")
fip_reservation_id=$(jq -rs 'map(select(.resource_type=="virtual:floatingip"))[].id' <<<"$reservations")

fip=$(openstack floating ip list --tags "reservation:$fip_reservation_id" -f json)
fip_id=$(jq -r '.[0]["ID"]' <<<"$fip")
fip_addr=$(jq -r '.[0]["Floating IP Address"]' <<<"$fip")

export FLOATING_IP_ID=$fip_id
export FLOATING_IP=$fip_addr
export RESERVATION_ID=$host_reservation_id
echo
echo "Stack parameters:"
echo " - floating_ip=$fip_id"
echo "   (exported as FLOATING_IP_ID)"
echo " - reservation_id=$host_reservation_id"
echo "   (exported as RESERVATION_ID)"
echo
echo "Also exported public IP address $fip_addr as FLOATING_IP"